"use strict";

this.name        = "torus-field-events.raeddi.alaric";
this.author      = "Alaric";
this.copyright   = "2016 Alaric";
this.description = "Raeddi Systems Ltd. Torus Field Monitor";
this.licence     = "CC BY-NC-SA 3.0 AU";



/**
 * Events:
 *
 *  This module generates additional events which your OXPs may hook into. Please
 *  see the this.$events object for the currently supported events. You can register
 *  for an event using code like this:
 *
 *		var events = this.worldScripts["torus-field-events.raeddi.alaric"];
 *		events.onTorusEngaged[this.name] = this.myEventHandler.bind(this);
 *
 *  To unregister the event:
 *
 *		delete events.onTorusEngaged[this.name];
  *
 *  All these events are triggered from within a frame callback. Your event handlers
 *  should be snappy and return ASAP. If you need to do extended processing, use
 *  timers.
 *
 *  onTorusEngaged(boolean)		Fires when player ship torus drive state changes
 *  onTorusCruise(boolean)		Fires when player ship crosses maxSpeed * 32 threshold
 *  onInjectorsEngaged(boolean)		Fires when player ship torus drive state changes
 *  onInjectorCruise(boolean)		Fires when player ship crosses maxSpeed * 7 threshold
 *  onPlayerTargetChanged
 *
 *
 *  extras?
 *  onAlertTCAS(entity, time)
 */

/**
 * Globals to track the last known state of player variables
 */
this.$torusEngaged = false;
this.$torusCruise = false;
this.$alertMassLocked = false;
this.$playerTarget = null;

this.$event = 
{
	onAlertMassLocked:		{ },
	onTorusEngaged:			{ },
	onTorusCruise:			{ },
	onPlayerTargetChanged:	{ }
};

this.$const =
{
	torusMultiplier:		32,
	injectorMultiplier:		7
};


/**
 * Frame callback to monitor several player variables for changes and call
 * event handlers when they do
 */
this._onFrameUpdate = function(delta)
{
	var torusCruise = player.ship.speed >= player.ship.maxSpeed * 32;
	var injectorCruise = player.ship.speed >= player.ship.maxSpeed * 7;

	if (this.$alertMassLocked !== player.ship.alertMassLocked)
	{
		this.$alertMassLocked = player.ship.alertMassLocked;
		this._onAlertMassLocked(this.$alertMassLocked);
	}

	if (this.$torusEngaged !== player.ship.torusEngaged)
	{
		this.$torusEngaged = player.ship.torusEngaged;
		this._onTorusEngaged(this.$torusEngaged);
	}

	if (this.$torusCruise !== torusCruise)
	{
		this.$torusCruise = torusCruise;
		this._onTorusCruise(this.$torusCruise);
	}

	if (this.$playerTarget !== player.ship.target)
	{
		this._onPlayerTargetChanged(this.$playerTarget, player.ship.target);
		this.$playerTarget = player.ship.target;
	}
}

this._onAlertMassLocked = function(boolean)
{
}

this._onTorusEngaged = function(boolean)
{
	for (var key in this.$event.onTorusEngaged)
	{
		log(this.name, "Calling onTorusEngaged(" + boolean + ") for " + key);
		this.$event.onTorusEngaged[key](boolean);
	}
}

/**
 * Note: onTorusEnagaged(false) will always occur before onTorusCruise(false)
 */
this._onTorusCruise = function(boolean)
{
	for (var key in this.$event.onTorusCruise)
	{
		log(this.name, "Calling onTorusCruise(" + boolean + ") for " + key);
		this.$event.onTorusCruise[key](boolean);
	}
}

/**
 * Fired when:
 *
 *  player.ship.injectorsEngaged is true and player.ship.speed reaches or
 *  exceeds 7 * player.ship.maxSpeed.
 *
 *  _onInjectorCruise(true) has been called and player.ship.speed falls
 *  below 7 * player.ship.maxSpeed.
 *
 *  If the player engages Torus drive whilst in injector cruise, the order
 *  of events is:
 *
 *	onInjectorsEngaged(true)
 *	onInjectorCruise(true)
 *	onTorusEngaged(true)
 *	onTorusCruise(true)
 *	onTorusEngaged(false)
 *	onTorusCruise(false)
 *	onInjectorsEngaged(false)
 *	onInjectorCruise(false)
 */
this._onInjectorCruise = function(boolean)
{
}

this._onPlayerTargetChanged = function(oldTarget, newTarget)
{
	for (var key in this.$event.onPlayerTargetChanged)
	{
		this.$event.onPlayerTargetChanged[key](oldTarget, newTarget);
	}
}


/**
 * Removes the frame callback _onFrameUpdate and deletes the handle
 */
this.shipWillDockWithStation = function(station)
{
	if (this.hasOwnProperty("$fcbHandle")) return;
	removeFrameCallback(this.$fcbHandle);
	delete this.$fcbHandle;
}

this.shipLaunchedFromStation = function(station)
{
	this.$torusEngaged = false;
	this.$torusCruise = false;
	this.$alertMassLocked = false;
	this.$playerTarget = null;


	for (var eventName in this.$event)
	{
		for (var eventHandler in this.$event[eventName])
		{
			log(this.name, eventName + " has a handler registered in " + eventHandler);
		}
	}

	/* do not install callback if it already exists (should never happen) */
	if (this.$fcbHandle != null && isValidFrameCallback(this.$fcbHandle))  return;
	this.$fcbHandle = addFrameCallback(this._onFrameUpdate.bind(this));
}