/**
 * Raeddi Torus Field Monitor (tm)
 */
"use strict";
this.name        = "torus-field-monitor.raeddi.alaric";
this.author      = "Alaric";
this.copyright   = "2016 Alaric";
this.description = "Raeddi Systems Ltd. Torus Field Monitor";
this.licence     = "CC BY-NC-SA 3.0 AU";

// Feel free to modify to taste.
this.$option =
{
	collisionWarningTime:  10.0,	// Seconds before impact of Torus auto-disengage (not yet implemented)
	collisionFailSafeTime:	2.0,	// Seconds before impact of Torus auto-disengage
	
	pingFrequency:		0.25, 			// seconds between polling all entities in system (only when torus engaged)


	/* controls when the "Torus Field Collapse" warning is pinged to the console. */
	// note: "duration" not yet implemented
	consolePingIntervals:
	[
		{ from: 120,	every: 30,	duration: 5.0 },
		{ from: 60,		every: 15,	duration: 5.0 },
		{ from: 30,		every: 10,	duration: 2.5 },
		{ from: 15,		every: 5,	duration: 2.0 },
		{ from: 10,		every: 1,	duration: 1.0 }
	]
};

/**
 * The following variables are public for read-only access. Other OXPs may
 * read these variables. I promise they will be available and valid in all
 * future versions.
 */
this.$public =
{
	/* intended for read-only access */
	timeToMassLock:			Infinity,	// Infinity if no lock possible, 0 if locked, or time in seconds.
	massLockZeroTime:		Infinity,	// Game clock adjusted time at predicted mass-lock
	numberOfDisturbances:	0,			// The number of Torus field disturbances currently being tracked.
	fieldSynchronised:		true,		// do not rely on this yet
	
	/* read/write variables */
	
	consolePingDisabled:	0			// increment to disable console messages, decrement when done
										// this ensures multiple disables by different scripts are synced.
};


/**
 * Other OXPs/scripts may access these if they desire, but they are not
 * guaranteed to be present or have the same properties or semantics in 
 * future releases.
 *
 * Mainly intended for access by other scripts within this OXP.
 */
this.$protected =
{
	disturbances:	[]					// see this._partitionLockingEntities for documentation
};

/**
 * Other scripts/OXPs should pretend these don't exist. Touching them may blow
 * a gasket... but I can't really do anything to stop you.
 */
this.$private =
{
	consolePingEnabled:		true,
	nextConsolePing:		null,
	signaturesByShipKey:	{},			// Cache of torus field signatures by shipdata key.
	potentialCollider:		null,
	collisionEntity:		null
};


// Loaded with worldscript objects if the corresponding OXP is found
// and verified to have the implementation details we are relying on.
this.$compatibility =
{
	variableMassLock: null
};


/**
 * The following 'friend' script keys will have their values loaded with
 * worldScript[key] when startUp runs.
 */
this.$friend =
{
	library:	"torus-field-events.raeddi.alaric",
	analyser:	"torus-field-analyser.raeddi.alaric",
	marker:		"torus-field-marker.raeddi.alaric"
};


// pseudo-constants
// These values should not be changed unless you have good reason.
this.$const =
{
	aquisitionTime:			1.0,						// seconds until target can be detected
	waypointKey:			"torus-field-monitor-oxp",
	beaconCode:				"m",						// m for mass
	beaconLabel:			"Torus Field Disturbance",
	torusMultiplier:		32,							// Torus drive speed multiplier.

	
	/* When Variable Mass Lock OXP is running, the following scan classes are
	 * considered non-ship entities, even when isShip returns true. This is
	 * used only for calculating the correct mass-lock range.
	 */
	nonVariableMassLock:			
	{
		CLASS_BUOY: true,
		CLASS_MINE: true,
		CLASS_MISSILE: true,
		CLASS_ROCK: true,
		CLASS_NO_DRAW: true
	}
};



/**
 * Returns the 'tag' object this script uses to attach information to an entity.
 * Keeping the information with the entity is much more conveinient than trying
 * to maintain some kind of mapping. Returns null if the entity is stale.
 */
this._getEntityTag = function(entity)
{
	var propertyName = "$" + this.name;

	if (!entity.isValid) return null;

	if (!entity.hasOwnProperty(propertyName))
	{
		/* create default object */
		entity[propertyName] =
		{
			aquisitionTime:			null,			// Clock time that the target will be aquired
			massLockRange:			null,			// Mass-lock range of entity in meters
			timeToMassLock:			Infinity,		// Time to mass lock in seconds			
			torusFieldSignature:	null,			// Number - constructed automagically
			torusFieldInfluence:	Infinity		// Maximum detectable mass-lock range in seconds
		};
	}

	return entity[propertyName];
};

/**
 * Using the $option.consolePingInterval table, returns the next console 
 * warning time if the current time is T
 */
this._getNextPingTime = function(T)
{
	var bounds = [0, this.$option.consolePingIntervals.length - 1];
	
	/* binary search of $option.consolePingIntervals for largest element <= T */	
	while (bounds[0] !== bounds[1])
	{
		var pivot = Math.floor((bounds[0] + bounds[1] + 1) / 2);
		var flag = (T > this.$option.consolePingIntervals[pivot].from) ? 1 : 0;
		bounds[flag] = pivot - flag;
	}

	var element = this.$option.consolePingIntervals[bounds[0]];

	//log (this.name, T + " (" + this.$option.consolePingIntervals[bounds[0]].from + ")");
	
	T = T - element.every - (T - element.from) % element.every;	

	if (++bounds[0] < this.$option.consolePingIntervals.length)
	{
		T = Math.max(T, this.$option.consolePingIntervals[bounds[0]].from);
	}

	return T;
};



/**
 * Determines the detectable range of a field disturbance, in seconds, given
 * the supplied disturbance signature. Note this range is measured in seconds
 * of Torus-drive travel time.
 *
 * Ie: An orbital shuttle has a range of about 6.8 seconds compared to over 45
 * seconds for an Icosolese station.
 *
 * When a field disturbance contains multiple ships, their signatures are
 * combined in a Pythagorean way (see _getSignatureForEntities). So 4 ships
 * of the same type will yeild double the range.
 */
this._getFieldDisturbanceRange = function(signature)
{
	return Math.pow(signature, Math.sqrt(0.5) / Math.PI) / (2.0 * Math.PI);
};


/**
 * Given a Torus Field signature, computes the "Influence" (detectable range)
 * of the signature.
 */
this._getInfluenceForSignature = function(signature)
{
	return Math.pow(signature, Math.sqrt(0.5) / Math.PI) / (2.0 * Math.PI);
};

/**
 * Generates a "signature" Number for the provided entity. The signature
 * is currently based upon a bit interleave of the mass and collisionRadius
 * properties. Larger and more massive objects have larger signatures.
 *
 * This value is cached after the first call on an entity. In addition,
 * signatures are also cached on shipDataKey. Calling on stale entities 
 * will return null.
 */
this._getSignatureForEntity = function(entity)
{
	var entityTag;
	
	if (!(entityTag = this._getEntityTag(entity))) return null;

	/* if the signature is already cached, don't calculate it again */
	if (entityTag.torusFieldSignature === null)
	{
		var signature = 0;
		
		/* check if we already have a signature for shipdatakey */
		if (entity.isShip && this.$private.signaturesByShipKey[entity.dataKey])
		{
			signature = this.$private.signaturesByShipKey[entity.dataKey];
		}
		else
		{
			var mass = entity.mass;
			var radius = Math.pow(entity.collisionRadius, 3);

			// determine bit entropy of mass and radius
			var massEntropy = (mass) ? Math.ceil(Math.log(mass) / Math.log(2)) : 0;
			var radiusEntropy = (radius) ? Math.ceil(Math.log(radius) / Math.log(2)) : 0;

			// get bit-shift required to globally justify to 15-bits
			var magnitude = Math.max(Math.max(massEntropy, radiusEntropy) - 15, 0);
	
			// justify the values to 15-bits
			mass = Math.floor(entity.mass / Math.pow(2, magnitude)) & 0x7fff;
			radius = Math.floor(radius / Math.pow(2, magnitude)) & 0x7fff;

			var shift = 0;
			var result = 0;

			for (var bit = 1; bit < 0x8000; bit <<= 1)
			{
				result |= (mass & bit) << shift;
				result |= (radius & bit) << ++shift;
			}

			// restore signature to pre-justification magnitude
			var signature = result * Math.pow(2, magnitude);
		}

		// record result in cache
		entityTag.torusFieldSignature = signature;
		entityTag.torusFieldInfluence = this._getInfluenceForSignature(signature);

		//log(this.name, entity.name + "|" + entity.scanClass + " torusFieldInfluence = " + entityTag.torusFieldInfluence);
		if (entity.isShip) this.$private.signaturesByShipKey[entity.dataKey] = signature;
	}
		
	return entityTag.torusFieldSignature;
};




/**
 * Input is an array of entities with non-zero non Infinity timeToMassLock. The
 * array must be sorted in order of increasing timeToMassLock.
 *
 * Returns an array of arrays of entities. Entities are grouped together according
 * to whether they will be considered part of the same Torus Field disturbance.
 *
 * Each sub-array contains an the following additional properties:
 *
 *	- $signature:		A computed value to allow us to track disturbance groups
 *						from one frame to the next. Not garanteed to be collision
 *						free. Depends on ship types and count in disturbance.
 *
 *	- $vector:			Player-relative Vector3D to weighted disturbance center.
 *
 *	- $timeToMassLock:	Time, in game clock seconds, until this disturbance
 *						will cause mass-lock. The major array is sorted based on
 *						this value, whith shortest TTML in [0].
 *
 * The first element of each sub-array is the main locking entity, with the other
 * being on the periphery.
 */
this._partitionLockingEntities = function(entities)
{
	var i;
	var roots = [];
	var minor;
	var major;

	for (minor = 0; minor < entities.length; ++minor)
	{
		var entityTag = this._getEntityTag(entities[minor]);
				
		for (major = 0; major < roots.length; ++major)
		{
			var D = entities[minor].position.distanceTo(roots[major][0].position);
			if (D <= entityTag.massLockRange && D <= roots[major].$massLockRange) break;
		}

		if (major === roots.length)
		{
			roots.push([]);
			roots[major].$timeToMassLock = entityTag.timeToMassLock;
			roots[major].$massLockRange = entityTag.massLockRange;
			roots[major].$signature = 0;
			roots[major].$weight = 0;
			roots[major].$vector = new Vector3D(0, 0, 0);			
		}
		
					
		roots[major].push(entities[minor]);		
		
		roots[major].$weight += entityTag.torusFieldSignature;
		roots[major].$signature += Math.pow(entityTag.torusFieldSignature, 2);
		
		/* add the relative vector, weighted by entity signature, to the disturbance vector */
		var D = entities[minor].position.subtract(player.ship.position);		
		roots[major].$vector = roots[major].$vector.add(D.multiply(entityTag.torusFieldSignature));
	}
	
	for (major = 0; major < roots.length; ++major)
	{
		
		roots[major].$signature = Math.sqrt(roots[major].$signature);		
		roots[major].$influence = this._getInfluenceForSignature(roots[major].$signature);
		roots[major].$vector = roots[major].$vector.multiply(1.0 / roots[major].$weight);
	}

	/* remove disturbances which are too distant to be detected */
	
	roots = roots.filter(
		function(element, index, array)
		{
			return element.$timeToMassLock <= element.$influence;
		}
	);

	return roots;
}




/**
 * This function will be called by the event monitor when torus drive has
 * entered into or exited cruise mode (32 * maxSpeed)
 */
this._onTorusCruise = function(cruising)
{
	if (cruising)
	{
		this.$pingTimer = new Timer(
			this,
			this._pingMassLock.bind(this),
			0,								// call immediately
			this.$option.pingFrequency		// then every pingFrequency seconds
		);	
	}	
	
	// if Torus has been disengaged (by player or mass-lock), delete the timer
	else if (this.$pingTimer)
	{
		// reset markers if mass-locked		
		if (player.alertMassLocked) this.$friend.marker._clearAllDisturbances();

		/* if TCAS has changed a scan class, restore it now */
		if (this.$private.collisionEntity !== null)
		{
			var entityTag = this._getEntityTag(this.$private.collisionEntity);
			if (entityTag) this.$private.collisionEntity.scanClass = entityTag.originalScanClass;
			this.$private.collisionEntity = null;
		}


		//this.$consolePing = false;
		this.$pingTimer.stop();					// stop timer first or Oolite will log unreferenced timer warning
		delete this.$pingTimer;	
		this.$protected.disturbances = [];		// clear all disturbances
				
	}
}

/**
 * If console ping is not disbaled (=== false), reset to true to indicate next disturbance
 * is 'fresh' (report "Disturbance Detected" message)
 */
this._resetConsolePinger = function()
{
	if (this.$private.consolePingEnabled !== false) this.$private.consolePingEnabled = true;
}

this._createMassLockTimer = function()
{
	if (this.$private.massLockTimer) return;

	this.$private.massLockTimer = addFrameCallback(this._massLockTimer.bind(this));
	this._resetConsolePinger();
}

/**
 * Delete the mass-lock timer frame callback if it exists
 */
this._deleteMassLockTimer = function()
{
	if (!this.$private.massLockTimer) return;
	
	removeFrameCallback(this.$private.massLockTimer);
	delete this.$private.massLockTimer;
	this._resetConsolePinger();
}

/**
 * Frame callback for updating this.$public.timeToMassLock
 */
this._massLockTimer = function(delta)
{
	/* The timer can be removed if no mass-lock can occur */
	if (this.$public.massLockZeroTime === Infinity)
	{
		this.$public.timeToMassLock = Infinity;
		this._deleteMassLockTimer();
		return;
	}

	/* Update timeToMassLock */
	this.$public.timeToMassLock = this.$public.massLockZeroTime - clock.adjustedSeconds;

	/* Delete timer if mass-lock has occurred */
	if (this.$public.timeToMassLock <= 0 || player.ship.alertMassLocked)
	{
		this.$public.massLockZeroTime = Infinity;
		this.$public.timeToMassLock = 0;
		this._deleteMassLockTimer();
		return;
	}

	/* return now if we are not actively pinging the console */
	if (this.$public.consolePingDisabled !== 0) return;
	
	/* do not ping unless we have crossed a integer second boundary */
	var pingTime = Math.ceil(this.$public.timeToMassLock);
	if (pingTime - this.$public.timeToMassLock > delta) return;

	/* ping if it is time to */
	if (this.$private.consolePingEnabled === true || pingTime <= this.$private.consolePingEnabled)
	{
		player.consoleMessage("Field collapse in " + pingTime + " seconds", 1);
		this.$private.consolePingEnabled = this._getNextPingTime(pingTime);
	}
	
}


/**
 * Returns the mass-lock range of all entities at Oolite 1.82. Will return 
 * Infinity if object is not mass-lockable. This value is cached after the
 * first call on an entity. Calling on stale entities will return null.
 */
this._entityMassLockRange = function(entity)
{
	var entityTag;

	if (!(entityTag = this._getEntityTag(entity))) return null;

	if (entityTag.massLockRange === null)
	{
		entityTag.massLockRange = 0;

		if (this._entityWillMassLock(entity))
		{
			// planets and stars mass-lock based upon their radii
			if (entity.isPlanet)
			{
				entityTag.massLockRange = Math.max(entity.collisionRadius * Math.sqrt(4), 25000);
			}
			else if (entity.isSun)
			{
				entityTag.massLockRange = Math.max(entity.collisionRadius * Math.sqrt(2), 25000);
			}
			else if (this.$compatibility.variableMassLock && entity.isShip)
			{
				if (this.$const.nonVariableMassLock[entity.scanClass])
				{
					entityTag.massLockRange = player.ship.scannerRange;
				}
				else
				{
					// VML only changes the mass-lock range of ships within scanner range
					entityTag.massLockRange = Math.min(
						this.$compatibility.variableMassLock.$Range(entity.mass),
						player.ship.scannerRange
						);						
				}
			}
			else
			{
				entityTag.massLockRange = player.ship.scannerRange;
			}
		}

		
	}

	return entityTag.massLockRange;
}



/**
 * Returns the _original_ scan class of the entity.
 *
 * This function is aware of the following OXPs:
 *	- Variable Mass Lock
 *
 */
this._getEntityScanClass = function(entity)
{
	var scanClass = entity.scanClass;

	if (entity === this.$private.collisionEntity)
	{
		var entityTag = this._getEntityTag(entity);
		scanClass = entityTag.originalScanClass;
	}

	/**
	 * If Variable Mass Lock is running, we need to watch out for CLASS_BUOY objects
	 * that are within scanner range, as VML may have changed their scan class to
	 * disable mass-lock.
	 */
	if (entity.scanClass === "CLASS_BUOY" && this.$compatibility.variableMassLock)
	{
		var variableMassLock = this.$compatibility.variableMassLock;

		if (variableMassLock.$VMActive)
		{
			/* try to find the entity in the VML array */
			var scanClassRecord = variableMassLock.$VMShips.filter(
				function(element, index, array)
				{
					return (element[0] === entity)
				}
			);

			/* use original scan class if found */
			if (scanClassRecord.length) scanClass = scanClassRecord[0][1];
		}
	}

	return scanClass;
}

/**
 * Returns true if the specified entity will cause mass-lock, false otherwise.
 *
 * TODO: This should be based on scanClass; but what to do with variableMassLock OXP?
 */
this._entityWillMassLock = function(entity)
{
	// Player never mass-locks themselves
	if (entity.isPlayer) return false;

	// Stars and planets always mass-lock	
	if (entity.isSun || entity.isPlanet) return true;		
	
	switch (this._getEntityScanClass(entity))
	{
		case "CLASS_BUOY":
		case "CLASS_CARGO":
		case "CLASS_ROCK":
		case "CLASS_NO_DRAW":
		case "CLASS_MINE":
			return false;

		default :
			return true;
	}
}


/**
 * Returns the number of seconds until <entity> intrudes to within <range>
 * of the player, if the player continues to travel along their current
 * heading at <velocity>. If <velocity> is omitted or is zero, the player
 * ships current velocity is used.
 *
 * This function ignores the entity velocity.
 */
this._entityTimeToIntrusion = function(entity, range, velocity)
{
	// calculate distance vector from player to entity
	var D = entity.position.subtract(player.ship.position);
	var distance = D.magnitude();

	/* if distance is less than range we are already intruded and the math below will NaN */
	if (distance < range) return 0;

	var cosTheta = Math.sqrt(1 - Math.pow(range / distance, 2));

	// cosine of angle between target direction and relative velocity
	var dotProduct = player.ship.heading.dot(D.direction());

	// if no intrusion possible
	if (dotProduct < cosTheta) return Infinity;

	// calculate distance to intrusion interface
	var Dintersect = distance * dotProduct - 
		Math.sqrt(
			 (Math.pow(dotProduct, 2) - 1) 
			* Math.pow(distance, 2)
			+ Math.pow(range, 2)
			);

	// return time to intrusion
	return Dintersect / ((velocity) ? velocity : player.ship.speed);
}


this._clearTCAS = function()
{
}

// return Infinity if no mass-lock will occur
// otherwise, returns number of seconds until predicted mass-lock
//
/**
 * Although this value is stored as .timeToMassLock under the entityTag, it is
 * recalculated anew every time this function is called. 
 */
this._entityTimeToMassLock = function(entity)
{
	if (!this._entityWillMassLock(entity))
	{
		if (entity.isPlayer) return Infinity;
		var T = this._entityTimeToIntrusion(entity, entity.collisionRadius + player.ship.collisionRadius, player.ship.speed);
		
		if (T < this.$option.collisionFailSafeTime && this.$private.collisionEntity === null) 
		{
			
			player.consoleMessage("TCAS Activated!", 5);
			var entityTag = this._getEntityTag(entity);
			this.$private.collisionEntity = entity;
			entityTag.originalScanClass = entity.scanClass;
			entity.scanClass = "CLASS_STATION";
		}
		return Infinity;
	}

	// I wish this was a property or method of Entity
	if (!this._entityWillMassLock(entity)) return Infinity;

	//log("_entityTimeToMassLock", entity.name);
	
	
	// calculate distance vector from player to entity
	var D = entity.position.subtract(player.ship.position);
	var Dmag = D.magnitude();	
	

	// Determine player torus velocity. If ship is travelling faster than default
	// Torus velocity, we will use that speed. Otherwise use default Torus. This
	// should provide enough compatibility with OXPs overriding Torus speed.
	var defaultTorusSpeed = player.ship.maxSpeed * 32;

	var Vplayer = (player.ship.speed > defaultTorusSpeed)
		? player.ship.velocity
		: player.ship.heading.multiply(player.ship.maxSpeed * 32)
		;


	// Experiments have shown it is best to ignore the entity velocity. It
	// is small enough to disregard, and maneuvering can result in incosistent
	// results between frames. We could, I guess, use an entities destination
	// as a guide, but that would be cheating.
	var Vrel = Vplayer;

	// get entity mass lock and disturbance range
	var Dlock = this._entityMassLockRange(entity);
	
			
	// if the entity is already within mass-lock range. Shouldn't happen, but
	// will result in NaNs if it does.
	if (Dmag < Dlock) 
	{
		
		return 0;
	}


	// cosine of maximum angle between entity and player ship heading which will
	// result in mass-lock
	//
	// If D is the distance to a point outside a sphere of radius R, then R/D
	// gives us the sine of the tangent angle (ie: apparent angular radius).
	//
	// sqrt(1 - sin(x) ^ 2) == cos(x)
	var cosTheta = Math.sqrt(1 - Math.pow(Dlock / Dmag, 2));

	// cosine of angle between target direction and relative velocity
	var dotProduct = player.ship.heading.dot(D.direction());
	
	var entityTag = this._getEntityTag(entity);
	
	// if no intersection no mass lock will occur
	if (dotProduct < cosTheta) 
	{
		entityTag.aquisitionTime = null;
		return Infinity;
	}

	if (entityTag.aquisitionTime === null)
	{
		entityTag.aquisitionTime = clock.adjustedSeconds + this.$const.aquisitionTime;		
		entityTag.isAquired = false;
		this.$public.fieldSynchronised = false;
		return Infinity;
	}

	if (clock.adjustedSeconds < entityTag.aquisitionTime) 
	{
		entityTag.isAquired = false;
		this.$public.fieldSynchronised = false;
		return Infinity;
	}

	entityTag.isAquired = true;

	// calculate distance to mass-lock interface
	var Dintersect = Dmag * dotProduct - 
		Math.sqrt(
			 (Math.pow(dotProduct, 2) - 1) 
			* Math.pow(Dmag, 2)
			+ Math.pow(Dlock, 2)
			);


	// return time to mass-lock
	return Dintersect / Vrel.magnitude();
}


this._setPrimaryDisturbance = function(disturbance)
{
	if (disturbance === null)
	{
		/* If scanner was previously clear, alert player */
		if (this.$public.massLockZeroTime !== Infinity)
		{
			if (this.$public.consolePingDisabled === 0)	player.consoleMessage("Torus field stablized", 3);
		}

		this.$public.timeToMassLock = Infinity;
		this.$public.massLockZeroTime = Infinity;
		this.$friend.marker._setPrimaryDisturbance(null);
		this._deleteMassLockTimer();
	}
	else
	{
		/* If scanner was previously clear, alert player */
		if (this.$public.massLockZeroTime === Infinity)
		{
			if (this.$public.consolePingDisabled === 0)	player.consoleMessage("Torus field disturbance detected", 1);
			var sound = new SoundSource;
			sound.sound = "[alert-condition-red]";
			sound.play();
		}		

		this.$public.timeToMassLock = disturbance.$timeToMassLock;
		this.$public.massLockZeroTime = clock.adjustedSeconds + this.$public.timeToMassLock;
		this.$friend.marker._setPrimaryDisturbance(disturbance.$vector);
		
		this._createMassLockTimer();
	}
}


/**
 * Polls all entities in the system. Use sparingly.
 *
 * Updates $public.timeToMassLock
 */
this._pingTorusField = function()
{
	/* cache the name of the entity tag property */
	var entityTagName = "$" + this.name;

	this.$public.fieldSynchronised = true;

	/* build array of all mass-locking entities in detectable range */
	var entities = system.filteredEntities(
		this,
		function (entity)
		{
			if (!entity.isValid) return false;

			var entityTag = this._getEntityTag(entity);
			this._getSignatureForEntity(entity);
						
			var T = this._entityTimeToMassLock(entity);
						
			entityTag.timeToMassLock = T;

			return (entityTag.timeToMassLock !== Infinity);
		}
	);

	/* sort in order of increasing time to mass-lock */
	entities.sort(
		function(lhs, rhs)
		{
			return lhs[entityTagName].timeToMassLock - rhs[entityTagName].timeToMassLock;
		}
	);

	/* group entities into disturbances */
	this.$protected.disturbances = this._partitionLockingEntities(entities);
	this.$public.numberOfDisturbances = this.$protected.disturbances.length;

	this._setPrimaryDisturbance(
		(this.$protected.disturbances.length > 0)
		? this.$protected.disturbances[0]
		: null
	);

}



/**
 * _pingMassLock is the timer-activated callback which polls the mass-lock state
 */
this._pingMassLock = function()
{

	/* if TCAS has changed a scan class, restore it now */
	if (this.$private.collisionEntity !== null)
	{
		var entityTag = this._getEntityTag(this.$private.collisionEntity);
		if (entityTag) this.$private.collisionEntity.scanClass = entityTag.originalScanClass;
		this.$private.collisionEntity = null;
	}

	// By desgin, this function should only be called when Torus is engaged
	if (!player.ship.torusEngaged)
	{
		log(this.name, "This is strange... _pingMassLock() was called when player.ship.torusEngaged === false.");
		return;
	}

	this._pingTorusField();

};

/**
 * When shipLaunchedFromStation, we install a frame callback to monitor the
 * state of the torus drive.
 */
this.shipLaunchedFromStation = function(station)
{
	
	

	var events = this.$friend.library.$event;
	events.onTorusCruise[this.name] = this._onTorusCruise.bind(this);
};


this.startUpComplete = function()
{
	this.$compatibility.variableMassLock = worldScripts["variablemasslock"];

	/* If variableMassLock is installed, make sure the implementation details we are relying
	 * on exist. If not, disable variableMassLock compatibility.
	 */
	if (this.$compatibility.variableMassLock)
	{
		var compatible = true;

		compatible = compatible && (typeof this.$compatibility.variableMassLock.$Range === "function");
		compatible = compatible && Array.isArray(this.$compatibility.variableMassLock.$VMShips);
		compatible = compatible && Array.isArray(this.$compatibility.variableMassLock.$VMRocks);

		if (!compatible)
		{
			log(this.name, "WARNING: Variable Mass Lock OXP is installed but is missing some expected properties.");
			log(this.name, "         Torus Field Monitor will bypass Variable Mass Lock.");
			this.$compatibility.variableMassLock = null;
		}
	}

	
	// report on enabled compatability modes
	log(this.name, "Enabled compatability modes:");
	var array = Object.keys(this.$compatibility);
	for (var i = 0; i < array.length; ++i)
	{
		log(this.name, "\t" + array[i] + ": " + !!this.$compatibility[array[i]]);
	}

	
};


this.startUp = function()
{
	for (key in this.$friend)
	{
		var name = this.$friend[key];
		var script = worldScripts[name];
		if (!script) log(this.name, "World script \"" + name + "\" is missing.");
		if (script) this.$friend[key] = script;
	}

};


/**
 * If shipWillDockWithStation, remove FCB.
 */
this.shipWillDockWithStation = function(station)
{
	this._deleteMassLockTimer();

	var events = worldScripts["torus-field-events.raeddi.alaric"].$event;
	delete events.onTorusCruise[this.name];


	// Kill our timer if it is still running
	if (this.$pingTimer)
	{
		this.$pingTimer.stop();
		delete this.$pingTimer;
	}
};


